from typing import Union

import numpy as np
import pandas as pd
from pint import UnitRegistry

from ..core import Comparison

UNITS = UnitRegistry()


class WeatherCodeComparison(Comparison):
    def __init__(
        self,
        observation: Union[np.ndarray, list],
        forecast: Union[np.ndarray, list],
    ):
        super().__init__(observation, forecast)
        self.df = pd.DataFrame(
            {
                "observation": self.observation,
                "forecast": self.forecast,
            }
        )

    def __repr__(self):
        return super().__repr__().replace("Comparison", "WeatherCodeComparison")

    def __str__(self):
        return super().__str__().replace("Comparison", "WeatherCodeComparison")

    def gather_all_factors(self):
        funcs = {
            # tuple: (function object, comment)
            "acc": (self.calc_multiclass_accuracy_ratio, "Accuracy"),
            "hss": (self.calc_multiclass_heidke_skill_score, "HSS"),
            "hk": (self.calc_multiclass_hanssen_kuipers_score, "HK"),
        }

        result = []
        for func_desc, (func, comment) in funcs.items():
            result.append(
                {
                    "element": "weather_code",
                    "indicator": func_desc,
                    "value": func(),
                    "comment": comment,
                }
            )


        df = pd.DataFrame(result)

        return df


if __name__ == "__main__":
    pass



"""
This is the module to store wind directions constants.

* DIRECTION8_CENTER_ANGLE: Wind 8 cardinal directions central angles.
* DIRECTION16_CENTER_ANGLE: Wind 16 cardinal directions central angles.
* DIRECTION8_ABBR: Wind 8 cardinal directions abbreviation.
* DIRECTION16_ABBR: Wind 16 cardinal directions abbreviation.
"""

DIRECTION8_CENTER_ANGLE = {0: 0, 1: 45, 2: 90, 3: 135, 4: 180, 5: 225, 6: 270, 7: 315}
DIRECTION16_CENTER_ANGLE = {
    0: 0,
    1: 22.5,
    2: 45,
    3: 67.5,
    4: 90,
    5: 112.5,
    6: 135,
    7: 157.5,
    8: 180,
    9: 202.5,
    10: 225,
    11: 247.5,
    12: 270,
    13: 292.5,
    14: 315,
    15: 337.5,
}

DIRECTION8_ABBR = {0: "N", 1: "NE", 2: "E", 3: "SE", 4: "S", 5: "SW", 6: "W", 7: "NW"}
DIRECTION16_ABBR = {
    0: "N",
    1: "NNE",
    2: "NE",
    3: "ENE",
    4: "E",
    5: "ESE",
    6: "SE",
    7: "SSE",
    8: "S",
    9: "SSW",
    10: "SW",
    11: "WSW",
    12: "W",
    13: "WNW",
    14: "NW",
    15: "NNW",
}



from typing import Union
from numbers import Number
from collections import Counter
from itertools import product

import numpy as np
from scipy import stats
import pandas as pd

from ..utils import (
    convert_to_ndarray,
    fix_zero_division,
    assert_length,
    drop_nan,
)
from .binarize import threshold_binarize


@assert_length
@drop_nan
def calc_binary_quadrant_values(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate binary quadrant values

    Quadrant values includes:
        * hits
        * misses
        * false_alarms
        * correct_rejects

    Args:
        observation (ndarray): Binarized obervation data array
                               that consist of True and False
        forecast (ndarray): Binarized forecast data array
                            that consist of True and False

    Returns:
        tuple: (hits, misses, false_alarms, correct_rejects, total)
    """
    hits = Counter(observation & forecast)[True]
    misses = Counter((observation ^ forecast) & observation)[True]
    false_alarms = Counter((observation ^ forecast) & forecast)[True]
    correct_rejects = Counter(~observation & ~forecast)[True]
    total = hits + misses + false_alarms + correct_rejects

    assert len(observation) == total

    return hits, misses, false_alarms, correct_rejects, total

@assert_length
@drop_nan
def calc_multiclass_confusion_matrix(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> pd.DataFrame:
    """Calculate multiclass confusion matrix

        confusion matrix
                            Observation
                            class 1     class 2      ...    class K
        Forecast    class 1 n(F_1,O_1)  n(F_1,O_2)          n(F_1,O_K)
                    class 2 n(F_2,O_1)  n(F_2,O_2)          n(F_2,O_K)
                    .
                    .
                    .
                    class K n(F_K,O_1)  n(F_K,O_2)          n(F_K,O_K)

        where n(Fi,Oj) denotes the number of forecasts in category i that had observations in category j and K is total class number

    Args:
        observation (Union[list, np.ndarray]): Multiclass obervation data array
                                                that consist of class labels.
        forecast (Union[list, np.ndarray]): Multiclass forecast data array
                                        that consist of class labels.

    Returns:
        pd.DataFrame: confusion matrix
    """

    cates = np.unique(np.concatenate([np.unique(observation), np.unique(forecast)]))
    confusion_matrix_list = []
    for obs_cate_, fcst_cate_ in product(cates, cates):
        count_cate_ = Counter((observation==obs_cate_) & (forecast==fcst_cate_))[True]
        confusion_matrix_list.append([obs_cate_, fcst_cate_, count_cate_])

    confusion_matrix = pd.DataFrame(np.array(confusion_matrix_list), columns=['observation','forecast', 'count'])
    confusion_matrix = confusion_matrix.pivot_table('count', index='forecast',columns='observation',aggfunc='sum').astype(int)

    assert len(observation) == np.sum(confusion_matrix.values)

    return confusion_matrix


@assert_length
@fix_zero_division
@drop_nan
def calc_multiclass_accuracy_ratio(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate accuracy of Multiclass observation and forecast.

       Accuray = \frac {\sum\limits_{i=1}^{K} n(F_i,O_i)} {Total} * 100

    Args:
        observation (Union[list, np.ndarray]): Multiclass observation data array
                                                that consist of class labels.
        forecast (Union[list, np.ndarray]): Multiclass forecast data array 
                                            that consist of class labels.

    Returns:
        float: The accuracy(%) of multiclass forecast. Perfect score 100.
    """

    confusion_matrix = calc_multiclass_confusion_matrix(
        observation, forecast
    )
    # compute the sum of hits of all categories
    all_hits = np.sum(confusion_matrix.values.diagonal())
    total = len(observation)

    return (all_hits  / total) * 100

@assert_length
@fix_zero_division
@drop_nan
def calc_multiclass_heidke_skill_score(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """calculate the Heidke Skill Score (HSS), which measures the 
    fraction of correct forecasts after eliminating those forecasts 
    which would be correct due purely to random chance. 

        HSS = \frac {\frac {1} {Total} \sum\limits_{i=1}^{K} n(F_i,O_i) - 
            \frac {1} {Total^2} \sum\limits_{i=1}^{K} N(F_i)N(O_i) } 
            {1 - \frac {1} {Total^2} \sum\limits_{i=1}^{K} N(F_i)*N(O_i)}

    Args:
        observation (Union[list, np.ndarray]): Multiclass observation data array
                                                that consist of class labels.
        forecast (Union[list, np.ndarray]): Multiclass forecast data array 
                                            that consist of class labels.

    Returns:
        float: HSS score. Perfect score 1.
    """

    confusion_matrix = calc_multiclass_confusion_matrix(
        observation, forecast
    )
    total = len(observation)

    # compute HSS score
    acc_ = np.sum(confusion_matrix.values.diagonal()) / total 
    reference_acc_ = np.sum(confusion_matrix.sum(axis=0).values * confusion_matrix.sum(axis=1).values) / (total**2)
    perfect_acc_ = 1
    hss_score_ = ( acc_ - reference_acc_ ) / (perfect_acc_ - reference_acc_)
    
    return hss_score_

@assert_length
@fix_zero_division
@drop_nan
def calc_multiclass_hanssen_kuipers_score(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """calculate the Hanssen and Kuipers Score (HSS), which is 
    similar to the Heidke skill score (above), except that in 
    the denominator the fraction of correct forecasts due to 
    random chance is for an unbiased forecast.

        HK = \frac {\frac {1} {Total} \sum\limits_{i=1}^{K} n(F_i,O_i) - 
            \frac {1} {Total^2} \sum\limits_{i=1}^{K} N(F_i)N(O_i) } 
            {1 - \frac {1} {Total^2} \sum\limits_{i=1}^{K} N(O_i)^2}

    Args:
        observation (Union[list, np.ndarray]): Multiclass observation data array
                                                that consist of class labels.
        forecast (Union[list, np.ndarray]): Multiclass forecast data array 
                                            that consist of class labels.

    Returns:
        float: HK score. Perfect score 1.
    """

    confusion_matrix = calc_multiclass_confusion_matrix(
        observation, forecast
    )
    total = len(observation)

    # compute HK score
    acc_ = np.sum(confusion_matrix.values.diagonal()) / total 
    reference_acc_ = np.sum(confusion_matrix.sum(axis=0).values * confusion_matrix.sum(axis=1).values) / (total**2)
    perfect_acc_ = 1
    unbias_reference_acc_ = np.sum(confusion_matrix.sum(axis=0).values**2) / (total**2)
    hk_score_ = ( acc_ - reference_acc_ ) / (perfect_acc_ - unbias_reference_acc_)
    
    return hk_score_

@assert_length
@fix_zero_division
@drop_nan
def calc_binary_accuracy_ratio(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate accuracy of binarized observation and forecast.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: The accuracy(%) of binarized forecast.
    """
    hits, _, _, correct_rejects, total = calc_binary_quadrant_values(
        observation, forecast
    )

    return ((hits + correct_rejects) / total) * 100


@assert_length
@fix_zero_division
@drop_nan
def calc_hit_ratio(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate hit ratio.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: Missing ratio(%)
    """
    hits, misses, _, correct_rejects, _ = calc_binary_quadrant_values(
        observation, forecast
    )

    return (hits / (hits + misses)) * 100


@assert_length
@fix_zero_division
@drop_nan
def calc_miss_ratio(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate missing ratio.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: Missing ratio(%)
    """
    hits, misses, _, correct_rejects, _ = calc_binary_quadrant_values(
        observation, forecast
    )

    return (misses / (hits + misses)) * 100


@assert_length
@fix_zero_division
@drop_nan
def calc_false_alarm_ratio(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate false alarm ratio, commonly known as FAR.

    Note that false alarm rate(POFD) is different from false alarm ratio(FAR).
    You can get more information about this from https://doi.org/10.1175/2009WAF2222300.1

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: False alarm ratio(%)
    """
    hits, _, false_alarms, _, _ = calc_binary_quadrant_values(observation, forecast)

    return (false_alarms / (hits + false_alarms)) * 100


@assert_length
@fix_zero_division
@drop_nan
def calc_false_alarm_rate(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate false alarm rate, commonly known as POFD.

    Note that false alarm rate(POFD) is different from false alarm ratio(FAR).
    You can get more information about this from https://doi.org/10.1175/2009WAF2222300.1

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: False alarm rate(%)
    """
    _, _, false_alarms, correct_rejects, _ = calc_binary_quadrant_values(
        observation, forecast
    )

    return (false_alarms / (correct_rejects + false_alarms)) * 100


@assert_length
@fix_zero_division
@drop_nan
def calc_ts(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate Threat score(TS)

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: Threat score(TS)
    """
    hits, misses, false_alarms, _, _ = calc_binary_quadrant_values(
        observation, forecast
    )

    return hits / (hits + false_alarms + misses)


@assert_length
@fix_zero_division
@drop_nan
def calc_ets(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate Equitable threat score(ETS)

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: Equitable threat score(ETS)
    """
    hits, misses, false_alarms, _, total = calc_binary_quadrant_values(
        observation, forecast
    )

    hits_ref = ((hits + misses) * (hits + false_alarms)) / total

    return (hits - hits_ref) / (hits + false_alarms + misses - hits_ref)


@assert_length
@fix_zero_division
@drop_nan
def calc_bias_score(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate Bias score

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of True and False.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of True and False.

    Returns:
        float: Bias score
    """
    hits, misses, false_alarms, _, _ = calc_binary_quadrant_values(
        observation, forecast
    )

    return (hits + false_alarms) / (hits + misses)


@assert_length
@fix_zero_division
@convert_to_ndarray
@drop_nan
def calc_diff_accuracy_ratio(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    limit: Union[float, int] = 1,
) -> float:
    """Calculate the accuracy within a limited difference scope.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        limit (Union[float, int]): The limited(or tolerable) scope between
                                   observation and forecast. Default to 1.

    Returns:
        float: The difference accuracy(%).
    """

    diff = np.abs(observation - forecast)

    bins = np.full_like(diff, False)
    bins[diff <= limit] = True

    total = len(bins)
    counter = Counter(bins)

    return counter[True] / total * 100


@assert_length
@fix_zero_division
@convert_to_ndarray
@drop_nan
def calc_rmse(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate RMSE(root-mean-square error).

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

    Returns:
        float: The RMSE of forecast to observation.
    """
    return np.sqrt(np.sum((observation - forecast) ** 2) / len(observation))


@assert_length
@fix_zero_division
@convert_to_ndarray
@drop_nan
def calc_mae(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate MAE(mean absolute error).

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

    Returns:
        float: The MAE of forecast to observation.
    """
    return np.sum(np.abs(observation - forecast)) / len(observation)


@assert_length
@fix_zero_division
@convert_to_ndarray
@drop_nan
def calc_mbe(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate MBE(mean bias error).

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

    Returns:
        float: The MBE of forecast to observation.
    """
    return np.sum(forecast - observation) / len(observation)


@assert_length
@fix_zero_division
@convert_to_ndarray
@drop_nan
def calc_rss(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate the residual sum of square between forecast and observation.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

    Returns:
        float: The residual sum of square of forecast and observation.
    """
    return np.sum((observation - forecast) ** 2)


@assert_length
@fix_zero_division
@convert_to_ndarray
@drop_nan
def calc_chi_square(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> float:
    """Calculate the χ2 between observation and forecast.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

    Returns:
        float: The chi square of forecast and observation.
    """
    return np.sum((observation - forecast) ** 2) / len(observation)


@assert_length
@convert_to_ndarray
@drop_nan
def calc_linregress(
    observation: Union[list, np.ndarray], forecast: Union[list, np.ndarray]
) -> tuple:
    """Calculste the parameters of linear regression.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

    Returns:
        tuple: Aguments of linear regression as follow:
               slope, intercept, r_value, p_value
    """
    linreg_result = stats.linregress(forecast, observation)[:4]

    return tuple(round(p, 4) for p in linreg_result)


@assert_length
@drop_nan
def calc_threshold_accuracy_ratio(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number,
    compare: str = ">=",
) -> float:
    """Calculate accuracy ratio of forecast filtered by threshold.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        threshold (Number): The threshold to filter obervation and forecast.
                            This parameter should be used with `compare` parameter,
                            The `compare` parameter will control the logical operator.
        compare (str, optional): The logical operator applying `threshold` parameter.
                                * '>' greater
                                * '<' less
                                * '>=' greater or equal
                                * '<=' less or equal
                                Defaults to '>='.

    Returns:
        float: The accuracy ratio of forecast filtered by threshold.
    """
    obs_bins, fct_bins = threshold_binarize(
        observation, forecast, threshold=threshold, compare=compare
    )

    return calc_binary_accuracy_ratio(obs_bins, fct_bins)


@assert_length
@drop_nan
def calc_threshold_hit_ratio(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number,
    compare: str = ">=",
) -> float:
    """Calculate hit ratio of forecast filtered by threshold.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        threshold (Number): The threshold to filter obervation and forecast.
                            This parameter should be used with `compare` parameter,
                            The `compare` parameter will control the logical operator.
        compare (str, optional): The logical operator applying `threshold` parameter.
                                * '>' greater
                                * '<' less
                                * '>=' greater or equal
                                * '<=' less or equal
                                Defaults to '>='.

    Returns:
        float: The missing ratio of forecast filtered by threshold.
    """
    obs_bins, fct_bins = threshold_binarize(
        observation, forecast, threshold=threshold, compare=compare
    )

    return calc_hit_ratio(obs_bins, fct_bins)


@assert_length
@drop_nan
def calc_threshold_miss_ratio(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number,
    compare: str = ">=",
) -> float:
    """Calculate missing ratio of forecast filtered by threshold.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        threshold (Number): The threshold to filter obervation and forecast.
                            This parameter should be used with `compare` parameter,
                            The `compare` parameter will control the logical operator.
        compare (str, optional): The logical operator applying `threshold` parameter.
                                * '>' greater
                                * '<' less
                                * '>=' greater or equal
                                * '<=' less or equal
                                Defaults to '>='.

    Returns:
        float: The missing ratio of forecast filtered by threshold.
    """
    obs_bins, fct_bins = threshold_binarize(
        observation, forecast, threshold=threshold, compare=compare
    )

    return calc_miss_ratio(obs_bins, fct_bins)


@assert_length
@drop_nan
def calc_threshold_false_alarm_ratio(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number,
    compare: str = ">=",
) -> float:
    """Calculate false alarm ratio of forecast filtered by threshold.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        threshold (Number): The threshold to filter obervation and forecast.
                            This parameter should be used with `compare` parameter,
                            The `compare` parameter will control the logical operator.
        compare (str, optional): The logical operator applying `threshold` parameter.
                                * '>' greater
                                * '<' less
                                * '>=' greater or equal
                                * '<=' less or equal
                                Defaults to '>='.

    Returns:
        float: The false alarm ratio of forecast filtered by threshold.
    """
    obs_bins, fct_bins = threshold_binarize(
        observation, forecast, threshold=threshold, compare=compare
    )

    return calc_false_alarm_ratio(obs_bins, fct_bins)


@assert_length
@drop_nan
def calc_threshold_bias_score(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number,
    compare: str = ">=",
) -> float:
    """Calculate bias score of forecast filtered by threshold.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        threshold (Number): The threshold to filter obervation and forecast.
                            This parameter should be used with `compare` parameter,
                            The `compare` parameter will control the logical operator.
        compare (str, optional): The logical operator applying `threshold` parameter.
                                * '>' greater
                                * '<' less
                                * '>=' greater or equal
                                * '<=' less or equal
                                Defaults to '>='.

    Returns:
        float: The bias score of forecast filtered by threshold.
    """
    obs_bins, fct_bins = threshold_binarize(
        observation, forecast, threshold=threshold, compare=compare
    )

    return calc_bias_score(obs_bins, fct_bins)


@assert_length
@drop_nan
def calc_threshold_ts(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number,
    compare: str = ">=",
) -> float:
    """Calculate Threat score(TS) of forecast filtered by threshold.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        threshold (Number): The threshold to filter obervation and forecast.
                            This parameter should be used with `compare` parameter,
                            The `compare` parameter will control the logical operator.
        compare (str, optional): The logical operator applying `threshold` parameter.
                                * '>' greater
                                * '<' less
                                * '>=' greater or equal
                                * '<=' less or equal
                                Defaults to '>='.

    Returns:
        float: The Threat score(TS) of forecast filtered by threshold.
    """
    obs_bins, fct_bins = threshold_binarize(
        observation, forecast, threshold=threshold, compare=compare
    )

    return calc_ts(obs_bins, fct_bins)


@assert_length
@drop_nan
def calc_threshold_mae(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number,
    compare: str = ">=",
) -> float:
    """Calculate MAE(mean absolute error) of forecast filtered by threshold.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.
        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.
        threshold (Number): The threshold to filter obervation and forecast.
                            This parameter should be used with `compare` parameter,
                            The `compare` parameter will control the logical operator.
        compare (str, optional): The logical operator applying `threshold` parameter.
                                * '>' greater
                                * '<' less
                                * '>=' greater or equal
                                * '<=' less or equal
                                Defaults to '>='.

    Returns:
        float: The MAE(mean absolute error) of forecast filtered by threshold.
    """
    obs_bins = np.full_like(observation, False)
    fct_bins = np.full_like(forecast, False)

    exec(f"obs_index = observation {compare} {threshold}")
    exec(f"fct_index = forecast {compare} {threshold}")

    obs_bins[locals()["obs_index"]] = True
    fct_bins[locals()["fct_index"]] = True

    return calc_mae(obs_bins, fct_bins)



import numpy as np

from cyeva.core.statistic import (
    calc_mae,
    calc_mbe,
    calc_threshold_hit_ratio,
    calc_threshold_false_alarm_ratio,
    calc_threshold_ts,
)


def test_calc_mae_1e3(benchmark):
    # thousand level
    obs = np.random.rand(int(1e3))
    fct = np.random.rand(int(1e3))

    def inner():
        calc_mae(obs, fct)

    benchmark(inner)


def test_calc_mae_1e6(benchmark):
    # million level
    obs = np.random.rand(int(1e6))
    fct = np.random.rand(int(1e6))

    def inner():
        calc_mae(obs, fct)

    benchmark(inner)


def test_calc_mae_1e7(benchmark):
    # ten million level
    obs = np.random.rand(int(1e7))
    fct = np.random.rand(int(1e7))

    def inner():
        calc_mae(obs, fct)

    benchmark(inner)


def test_calc_mbe_1e3(benchmark):
    # thousand level
    obs = np.random.rand(int(1e3))
    fct = np.random.rand(int(1e3))

    def inner():
        calc_mbe(obs, fct)

    benchmark(inner)


def test_calc_mbe_1e6(benchmark):
    # million level
    obs = np.random.rand(int(1e6))
    fct = np.random.rand(int(1e6))

    def inner():
        calc_mbe(obs, fct)

    benchmark(inner)


def test_calc_mbe_1e7(benchmark):
    # ten million level
    obs = np.random.rand(int(1e7))
    fct = np.random.rand(int(1e7))

    def inner():
        calc_mbe(obs, fct)

    benchmark(inner)


def test_calc_threshold_hit_ratio_1e3(benchmark):
    # thousand level
    obs = np.random.rand(int(1e3))
    fct = np.random.rand(int(1e3))

    def inner():
        calc_threshold_hit_ratio(obs, fct, 0.5)

    benchmark(inner)


def test_calc_threshold_hit_ratio_1e6(benchmark):
    # million level
    obs = np.random.rand(int(1e6))
    fct = np.random.rand(int(1e6))

    def inner():
        calc_threshold_hit_ratio(obs, fct, 0.5)

    benchmark(inner)


def test_calc_threshold_hit_ratio_1e7(benchmark):
    # ten million level
    obs = np.random.rand(int(1e7))
    fct = np.random.rand(int(1e7))

    def inner():
        calc_threshold_hit_ratio(obs, fct, 0.5)

    benchmark(inner)


def test_calc_threshold_false_alarm_ratio_1e3(benchmark):
    obs = np.random.rand(int(1e3))
    fct = np.random.rand(int(1e3))

    def inner():
        calc_threshold_false_alarm_ratio(obs, fct, 0.5)

    benchmark(inner)


def test_calc_threshold_false_alarm_ratio_1e6(benchmark):
    # million level
    obs = np.random.rand(int(1e6))
    fct = np.random.rand(int(1e6))

    def inner():
        calc_threshold_false_alarm_ratio(obs, fct, 0.5)

    benchmark(inner)


def test_calc_threshold_false_alarm_ratio_1e7(benchmark):
    # ten million level
    obs = np.random.rand(int(1e7))
    fct = np.random.rand(int(1e7))

    def inner():
        calc_threshold_false_alarm_ratio(obs, fct, 0.5)

    benchmark(inner)


def test_calc_threshold_ts_1e3(benchmark):
    # thousand level
    obs = np.random.rand(int(1e3))
    fct = np.random.rand(int(1e3))

    def inner():
        calc_threshold_ts(obs, fct, 0.5)

    benchmark(inner)


def test_calc_threshold_ts_1e6(benchmark):
    # million level
    obs = np.random.rand(int(1e6))
    fct = np.random.rand(int(1e6))

    def inner():
        calc_threshold_ts(obs, fct, 0.5)

    benchmark(inner)


def test_calc_threshold_ts_1e7(benchmark):
    # ten million level
    obs = np.random.rand(int(1e7))
    fct = np.random.rand(int(1e7))

    def inner():
        calc_threshold_ts(obs, fct, 0.5)

    benchmark(inner)



"""
This is the module to store precipitation's level constants.

* PRECIP_LEVELS: Precipitation's levels thresholds for 1 hour, 3 hours and 24 hours.
* ACC_PRECIP_LEVELS: Accumulated precipitation's levels thresholds for 1 hour,
                     3 hours and 24 hours.
"""

import numpy as np

PRECIP_LEVELS = {
    "1h": {
        1: {"min": 0.1, "max": 1.9, "text": "light_rain"},
        2: {"min": 2, "max": 4.9, "text": "moderate_rain"},
        3: {"min": 5, "max": 9.9, "text": "heavy_rain"},
        4: {"min": 10, "max": 19.9, "text": "violent_rain"},
        5: {"min": 20, "max": np.inf, "text": "severe_violent_rain"},
    },
    "3h": {
        1: {"min": 0.1, "max": 2.9, "text": "light_rain"},
        2: {"min": 3, "max": 9.9, "text": "moderate_rain"},
        3: {"min": 10, "max": 19.9, "text": "heavy_rain"},
        4: {"min": 20, "max": 49.9, "text": "violent_rain"},
        5: {"min": 50, "max": 69.9, "text": "severe_violent_rain"},
        6: {"min": 70, "max": np.inf, "text": "super_severe_violent_rain"},
    },
    "12h": {
        1: {"min": 0.1, "max": 4.9, "text": "light_rain"},
        2: {"min": 5, "max": 14.9, "text": "moderate_rain"},
        3: {"min": 15, "max": 29.9, "text": "heavy_rain"},
        4: {"min": 30, "max": 69.9, "text": "violent_rain"},
        5: {"min": 70, "max": 139.9, "text": "severe_violent_rain"},
        6: {"min": 140, "max": np.inf, "text": "super_severe_violent_rain"},
    },
    "24h": {
        1: {"min": 0.1, "max": 9.9, "text": "light_rain"},
        2: {"min": 10, "max": 24.9, "text": "moderate_rain"},
        3: {"min": 25, "max": 49.9, "text": "heavy_rain"},
        4: {"min": 50, "max": 99.9, "text": "violent_rain"},
        5: {"min": 100, "max": 249.9, "text": "severe_violent_rain"},
        6: {"min": 250, "max": np.inf, "text": "super_severe_violent_rain"},
    },
}

ACC_PRECIP_LEVELS = {
    "1h": {
        1: {"min": 0.1, "max": np.inf, "text": "light_rain"},
        2: {"min": 2, "max": np.inf, "text": "moderate_rain"},
        3: {"min": 5, "max": np.inf, "text": "heavy_rain"},
        4: {"min": 10, "max": np.inf, "text": "violent_rain"},
        5: {"min": 20, "max": np.inf, "text": "severe_violent_rain"},
    },
    "3h": {
        1: {"min": 0.1, "max": np.inf, "text": "light_rain"},
        2: {"min": 3, "max": np.inf, "text": "moderate_rain"},
        3: {"min": 10, "max": np.inf, "text": "heavy_rain"},
        4: {"min": 20, "max": np.inf, "text": "violent_rain"},
        5: {"min": 50, "max": np.inf, "text": "severe_violent_rain"},
        6: {"min": 70, "max": np.inf, "text": "super_severe_violent_rain"},
    },
    "12h": {
        1: {"min": 0.1, "max": np.inf, "text": "light_rain"},
        2: {"min": 5, "max": np.inf, "text": "moderate_rain"},
        3: {"min": 15, "max": np.inf, "text": "heavy_rain"},
        4: {"min": 30, "max": np.inf, "text": "violent_rain"},
        5: {"min": 70, "max": np.inf, "text": "severe_violent_rain"},
        6: {"min": 140, "max": np.inf, "text": "super_severe_violent_rain"},
    },
    "24h": {
        1: {"min": 0.1, "max": np.inf, "text": "light_rain"},
        2: {"min": 10, "max": np.inf, "text": "moderate_rain"},
        3: {"min": 25, "max": np.inf, "text": "heavy_rain"},
        4: {"min": 50, "max": np.inf, "text": "violent_rain"},
        5: {"min": 100, "max": np.inf, "text": "severe_violent_rain"},
        6: {"min": 250, "max": np.inf, "text": "super_severe_violent_rain"},
    },
}



from typing import Union

import numpy as np
import pandas as pd
from pint import UnitRegistry

from ..core import Comparison

UNITS = UnitRegistry()


class TemperatureComparison(Comparison):
    def __init__(
        self,
        observation: Union[np.ndarray, list],
        forecast: Union[np.ndarray, list],
        unit: str = "degC",
        kind: str = None,
        lev: str = None,
    ):
        super().__init__(observation, forecast)
        self.kind = kind
        self.lev = lev
        self.observation = (self.observation * UNITS.parse_expression(unit)).to("degC").magnitude
        self.forecast = (self.forecast * UNITS.parse_expression(unit)).to("degC").magnitude
        self.df = pd.DataFrame(
            {
                "observation": self.observation,
                "forecast": self.forecast,
            }
        )

    def __repr__(self):
        return super().__repr__().replace("Comparison", "TemperatureComparison(degC)")

    def __str__(self):
        return super().__str__().replace("Comparison", "TemperatureComparison(degC)")

    def gather_all_factors(self):
        LEVS = {
            1: ("limit", "within 1 degC"),
            2: ("limit", "within 2 degC"),
        }
        funcs = {
            # tuple: (function object, unit, comment)
            "rmse": (self.calc_rmse, "degC", "RMSE"),
            "mae": (self.calc_mae, "degC", "MAE"),
            "rss": (
                self.calc_rss,
                "degC ** 2",
                None,
            ),
            "chi_square": (self.calc_chi_square, "degC ** 2", "χ2"),
        }

        result = []
        for func_desc, (func, units, comment) in funcs.items():
            result.append(
                {
                    "element": "temperature",
                    "interval": np.nan,
                    "kind": np.nan,
                    "level": np.nan,
                    "indicator": func_desc,
                    "value": func(),
                    "unit": units,
                    "comment": comment,
                }
            )

        for lev, (lev_kind, lev_desc) in LEVS.items():
            result.append(
                {
                    "element": "temperature",
                    "interval": np.nan,
                    "kind": lev_kind,
                    "level": lev_desc,
                    "indicator": "accuracy_ratio",
                    "value": self.calc_diff_accuracy_ratio(limit=lev),
                    "unit": "%",
                    "comment": np.nan,
                }
            )

        df = pd.DataFrame(result)

        return df


if __name__ == "__main__":
    pass



from typing import Union
from numbers import Number

import numpy as np
import pandas as pd
from pint import Quantity

from ..utils import result_round_digit, source_round_digit
from .binarize import threshold_binarize
from .statistic import (
    calc_diff_accuracy_ratio,
    calc_rmse,
    calc_mae,
    calc_mbe,
    calc_rss,
    calc_chi_square,
    calc_linregress,
    calc_bias_score,
    calc_binary_accuracy_ratio,
    calc_threshold_accuracy_ratio,
    calc_threshold_hit_ratio,
    calc_threshold_miss_ratio,
    calc_threshold_false_alarm_ratio,
    calc_threshold_bias_score,
    calc_threshold_ts,
    calc_threshold_mae,
    calc_multiclass_accuracy_ratio,
    calc_multiclass_hanssen_kuipers_score,
    calc_multiclass_heidke_skill_score
)


class Comparison:
    """A base class of comparison between obervation & forecast"""

    def __init__(
        self, observation: Union[np.ndarray, list], forecast: Union[np.ndarray, list]
    ):

        if isinstance(observation, Quantity):
            observation = observation.magnitude
        if isinstance(forecast, Quantity):
            forecast = forecast.magnitude

        self.df = pd.DataFrame({"observation": observation, "forecast": forecast})
        self.observation = self.df["observation"].values
        self.forecast = self.df["forecast"].values

    def __repr__(self):
        text = "\n".join(["<Object:Comparison>", "data:", self.df.__str__()])
        return text

    def __str__(self):
        text = "\n".join(["<Object:Comparison>", "data:", self.df.__str__()])
        return text

    @result_round_digit(4)
    @source_round_digit()
    def calc_rmse(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        *args,
        **kwargs
    ) -> float:
        "Root Mean Square Error"
        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_rmse(observation, forecast)

    @result_round_digit(4)
    @source_round_digit()
    def calc_mae(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        *args,
        **kwargs
    ) -> float:
        """Mean Absolute Error"""
        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_mae(observation, forecast)

    @result_round_digit(4)
    @source_round_digit()
    def calc_mbe(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        *args,
        **kwargs
    ) -> float:
        """Mean Bias Error"""
        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_mbe(observation, forecast)

    @result_round_digit(4)
    @source_round_digit()
    def calc_chi_square(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        *args,
        **kwargs
    ) -> float:
        """χ2"""
        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_chi_square(observation, forecast)

    @result_round_digit(4)
    @source_round_digit()
    def calc_rss(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        *args,
        **kwargs
    ) -> float:
        """Residual Sum of Square"""
        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_rss(observation, forecast)

    def calc_linregress_args(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        *args,
        **kwargs
    ):
        """linregress args"""
        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_linregress(observation, forecast)

    @result_round_digit(4)
    @source_round_digit()
    def calc_bias(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        threshold: Number = 0,
        *args,
        **kwargs
    ) -> float:
        """BIAS score"""
        if observation is None:
            _observation = self.observation
        else:
            _observation = observation
        if forecast is None:
            _forecast = self.forecast
        else:
            _forecast = forecast

        observation, forecast = threshold_binarize(
            _observation, _forecast, threshold=threshold, compare=">"
        )
        return calc_bias_score(observation, forecast)

    @result_round_digit(4)
    @source_round_digit()
    def calc_binary_accuracy_ratio(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        threshold: Number = 0,
        *args,
        **kwargs
    ) -> float:
        """Binary accuracy ratio

        Args:
            threshold (Number, optional): The threshold to binarize array.
                                                     Defaults to 0.

        Returns:
            float: The accuracy ratio.
        """
        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        binary_observation, binary_forecast = threshold_binarize(
            observation, forecast, threshold=threshold, compare=">"
        )
        return calc_binary_accuracy_ratio(binary_observation, binary_forecast)

    @result_round_digit(4)
    def calc_multiclass_accuracy_ratio(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        *args,
        **kwargs
    ) -> float:
        
        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_multiclass_accuracy_ratio(observation, forecast)

    @result_round_digit(4)
    def calc_multiclass_hanssen_kuipers_score(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        *args,
        **kwargs
    ) -> float:
        
        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_multiclass_hanssen_kuipers_score(observation, forecast)
    
    @result_round_digit(4)
    def calc_multiclass_heidke_skill_score(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        *args,
        **kwargs
    ) -> float:
        
        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_multiclass_heidke_skill_score(observation, forecast)

    @result_round_digit(4)
    @source_round_digit()
    def calc_diff_accuracy_ratio(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        limit: Number = 1,
        *args,
        **kwargs
    ) -> float:
        """
        Calculate the accuracy ratio that
        two array's difference within a limit scope.

        Args:
            limit (Number, optional): The limit of difference.
                                                 Defaults to 1.

        Returns:
            float: The accuracy ratio.
        """
        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_diff_accuracy_ratio(observation, forecast, limit=limit)

    @result_round_digit(4)
    @source_round_digit()
    def calc_threshold_accuracy_ratio(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        threshold: Number = 1,
        compare: str = ">=",
        *args,
        **kwargs
    ) -> float:

        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_threshold_accuracy_ratio(
            observation, forecast, threshold=threshold, compare=compare
        )

    @result_round_digit(4)
    @source_round_digit()
    def calc_threshold_hit_ratio(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        threshold: Number = 1,
        compare: str = ">=",
        *args,
        **kwargs
    ) -> float:

        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_threshold_hit_ratio(
            observation, forecast, threshold=threshold, compare=compare
        )

    @result_round_digit(4)
    @source_round_digit()
    def calc_threshold_miss_ratio(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        threshold: Number = 1,
        compare: str = ">=",
        *args,
        **kwargs
    ) -> float:

        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_threshold_miss_ratio(
            observation, forecast, threshold=threshold, compare=compare
        )

    @result_round_digit(4)
    @source_round_digit()
    def calc_threshold_false_alarm_ratio(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        threshold: Number = 1,
        compare: str = ">=",
        *args,
        **kwargs
    ) -> float:

        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_threshold_false_alarm_ratio(
            observation, forecast, threshold=threshold, compare=compare
        )

    @result_round_digit(4)
    @source_round_digit()
    def calc_threshold_bias_score(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        threshold: Number = 1,
        compare: str = ">=",
        *args,
        **kwargs
    ) -> float:

        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_threshold_bias_score(
            observation, forecast, threshold=threshold, compare=compare
        )

    @result_round_digit(4)
    @source_round_digit()
    def calc_threshold_ts(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        threshold: Number = 1,
        compare: str = ">=",
        *args,
        **kwargs
    ) -> float:

        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_threshold_ts(
            observation, forecast, threshold=threshold, compare=compare
        )

    @result_round_digit(4)
    @source_round_digit()
    def calc_threshold_mae(
        self,
        observation: Union[np.ndarray, list] = None,
        forecast: Union[np.ndarray, list] = None,
        threshold: Number = 1,
        compare: str = ">=",
        *args,
        **kwargs
    ) -> float:

        if observation is None:
            observation = self.observation
        if forecast is None:
            forecast = self.forecast

        return calc_threshold_mae(
            observation, forecast, threshold=threshold, compare=compare
        )



from typing import Union
from numbers import Number
from functools import partial

import numpy as np
import pandas as pd
from pint import UnitRegistry

from ..config.levels.precip import PRECIP_LEVELS, ACC_PRECIP_LEVELS
from ..core.binarize import (
    threshold_binarize,
    level_binarize,
)
from ..utils import result_round_digit, source_round_digit
from .statistic import (
    calc_binary_accuracy_ratio,
    calc_hit_ratio,
    calc_miss_ratio,
    calc_false_alarm_ratio,
    calc_false_alarm_rate,
    calc_ts,
    calc_ets,
    calc_bias_score,
)
from ..core import Comparison

UNITS = UnitRegistry()


@result_round_digit(2)
@source_round_digit(2)
def calc_precip_occur_indicators(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    threshold: Number = 0.0001,
    indicator: str = "accuracy_ratio",
    *args,
    **kwargs,
):
    """Calculate binarized indicators of precipitation occurrence

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.

        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

        threshold (Number): The threshold to filter obervation and forecast.
                            Values less than threshold will be dropped.
                            Default to 0.0001.

        indicator (str): Indicator name, Options:
                         * accuracy_ratio: binarized accuracy ratio
                         * miss_ratio: binarized missing ratio
                         * false_alarm_ratio: binarized false alarm ratio
                         * ts: binarized ts
                         * ets: binarized ets
                         * bias_score: binarized bias score

    Returns:
        float: Binarized indicators of precipitation occurrence(%)
    """
    observation, forecast = threshold_binarize(
        observation, forecast, threshold=threshold
    )
    indicator_funcs = {
        "accuracy_ratio": calc_binary_accuracy_ratio,
        "hit_ratio": calc_hit_ratio,
        "miss_ratio": calc_miss_ratio,
        "false_alarm_ratio": calc_false_alarm_ratio,
        "false_alarm_rate": calc_false_alarm_rate,
        "ts": calc_ts,
        "ets": calc_ets,
        "bias_score": calc_bias_score,
    }
    return indicator_funcs[indicator](observation, forecast)


@result_round_digit(2)
@source_round_digit(2)
def calc_precip_interval_indicators(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    kind: str,
    lev: int,
    indicator: str = "accuracy_ratio",
    *args,
    **kwargs,
) -> float:
    """Calculate indicators of specific level of precipitation.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.

        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

        kind (str): Statistic kind. Options are '1h', '3h', '12h' and '24h'.

        lev (int): Level ID, Options:
                   * 1: Light rain
                   * 2: Moderate rain
                   * 3: Heavy rain
                   * 4: Violent rain
                   * 5: Severe violent rain
                   * 6: Super severe violent rain(1h don't have this)

        indicator (str): Indicator name, Options:
                         * accuracy_ratio: binarized accuracy ratio
                         * miss_ratio: binarized missing ratio
                         * false_alarm_ratio: binarized false alarm ratio
                         * ts: binarized ts
                         * ets: binarized ets
                         * bias_score: binarized bias score

    Returns:
        float: The indicators in specific level
    """
    observation, forecast = level_binarize(
        observation, forecast, PRECIP_LEVELS[kind], lev
    )
    indicator_funcs = {
        "accuracy_ratio": calc_binary_accuracy_ratio,
        "hit_ratio": calc_hit_ratio,
        "miss_ratio": calc_miss_ratio,
        "false_alarm_ratio": calc_false_alarm_ratio,
        "false_alarm_rate": calc_false_alarm_rate,
        "ts": calc_ts,
        "ets": calc_ets,
        "bias_score": calc_bias_score,
    }

    return indicator_funcs[indicator](observation, forecast)


@result_round_digit(2)
@source_round_digit(2)
def calc_precip_accumulate_indicators(
    observation: Union[list, np.ndarray],
    forecast: Union[list, np.ndarray],
    kind: str,
    lev: int,
    indicator: str = "accuracy_ratio",
    *args,
    **kwargs,
):
    """Calculate indicators of specific level of precipitation.

    Args:
        observation (Union[list, np.ndarray]): Binarized observation data array
                                               that consist of numbers.

        forecast (Union[list, np.ndarray]): Binarized forecast data array that
                                            consist of numbers.

        kind (str): Statistic kind. Options are '1h', '3h', '12h' and '24h'.

        lev (int): Level ID, Options:
                   * 1: Light rain
                   * 2: Moderate rain
                   * 3: Heavy rain
                   * 4: Violent rain
                   * 5: Severe violent rain
                   * 6: Super severe violent rain(1h don't have this)

        indicator (str): Indicator name, Options:
                         * accuracy_ratio: binarized accuracy ratio
                         * miss_ratio: binarized missing ratio
                         * false_alarm_ratio: binarized false alarm ratio
                         * ts: binarized ts
                         * ets: binarized ets
                         * bias_score: binarized bias score

    Returns:
        float: The indicators in specific level
    """
    observation, forecast = level_binarize(
        observation, forecast, ACC_PRECIP_LEVELS[kind], lev
    )

    indicator_funcs = {
        "accuracy_ratio": calc_binary_accuracy_ratio,
        "hit_ratio": calc_hit_ratio,
        "miss_ratio": calc_miss_ratio,
        "false_alarm_ratio": calc_false_alarm_ratio,
        "false_alarm_rate": calc_false_alarm_rate,
        "ts": calc_ts,
        "ets": calc_ets,
        "bias_score": calc_bias_score,
    }

    return indicator_funcs[indicator](observation, forecast)


class PrecipitationComparison(Comparison):
    """Comparison of Precipitation"""

    def __init__(
        self,
        observation: Union[np.ndarray, list],
        forecast: Union[np.ndarray, list],
        unit: str = "mm",
        kind: str = None,
        lev: str = None,
    ):
        super().__init__(observation, forecast)
        self.kind = kind
        self.unit = unit
        self.lev = lev
        self.observation = (self.observation * UNITS.parse_expression(unit)).to("mm").magnitude
        self.forecast = (self.forecast * UNITS.parse_expression(unit)).to("mm").magnitude
        self.df = pd.DataFrame(
            {
                "observation": self.observation,
                "forecast": self.forecast,
            }
        )

        self.lev_index = [
            "0",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "+1",
            "+2",
            "+3",
            "+4",
            "+5",
            "+6",
        ]

    def __repr__(self):
        return super().__repr__().replace("Comparison", "PrecipitationComparison(mm)")

    def __str__(self):
        return super().__str__().replace("Comparison", "PrecipitationComparison(mm)")

    @staticmethod
    def __make_func_choice(indicator="accuracy_ratio"):
        interval_choice = dict.fromkeys(
            ["1", "2", "3", "4", "5", "6"],
            partial(calc_precip_interval_indicators, indicator=indicator),
        )
        accumulate_choice = dict.fromkeys(
            ["+1", "+2", "+3", "+4", "+5", "+6"],
            partial(calc_precip_accumulate_indicators, indicator=indicator),
        )

        func_choice = {"0": partial(calc_precip_occur_indicators, indicator=indicator)}
        func_choice.update(interval_choice)
        func_choice.update(accumulate_choice)

        return func_choice

    def calc_accuracy_ratio(self, kind: str = None, lev: str = "0"):
        """Calculste precipitation accuracy ratio.

        Args:
            kind (str): Statistic kind. Options are '1h', '3h', '12h' and '24h'.
            lev (str): Level ID, Options:
                       * "0": Rain or shine.
                       * "1": Light rain.
                       * "2": Moderate rain.
                       * "3": Heavy rain.
                       * "4": Violent rain.
                       * "5": Severe violent rain.
                       * "6": Super severe violent rain(1h don't have this).
                       * "+1": Accumulated to Light rain.
                       * "+2": Accumulated to Moderate rain.
                       * "+3": Accumulated to Heavy rain.
                       * "+4": Accumulated to Violent rain.
                       * "+5": Accumulated to Severe violent rain.
                       * "+6": Accumulated to Super severe violent rain(1h don't have this).
                       Defaults to "0".

        Returns:
            float: Accuracy ratio of specific kind and level(%).
        """
        if int(lev) > 0:
            if not kind:
                kind = self.kind
            assert kind and (kind.lower() in ["1h", "3h", "12h", "24h"])
            if not lev:
                lev = self.lev
            assert lev and (lev in self.lev_index)

            if kind == "1h" and lev in ["6", "+6"]:
                lev = lev.replace("6", "5")

            kind = kind.lower()
        else:
            kind = None

        func_choice = self.__make_func_choice(indicator="accuracy_ratio")

        return func_choice[lev](
            self.observation, self.forecast, kind=kind, lev=int(lev)
        )

    def calc_hit_ratio(self, kind: str = None, lev: str = "0"):
        """Calculste precipitation hit ratio.

        Args:
            kind (str): Statistic kind. Options are '1h', '3h', '12h' and '24h'.
            lev (str): Level ID, Options:
                       * "0": Rain or shine.
                       * "1": Light rain.
                       * "2": Moderate rain.
                       * "3": Heavy rain.
                       * "4": Violent rain.
                       * "5": Severe violent rain.
                       * "6": Super severe violent rain(1h don't have this).
                       * "+1": Accumulated to Light rain.
                       * "+2": Accumulated to Moderate rain.
                       * "+3": Accumulated to Heavy rain.
                       * "+4": Accumulated to Violent rain.
                       * "+5": Accumulated to Severe violent rain.
                       * "+6": Accumulated to Super severe violent rain(1h don't have this).
                       Defaults to "0".

        Returns:
            float: Hit ratio of specific kind and level(%).
        """
        if int(lev) > 0:
            if not kind:
                kind = self.kind
            assert kind and (kind.lower() in ["1h", "3h", "12h", "24h"])
            if not lev:
                lev = self.lev
            assert lev and (lev in self.lev_index)

            if kind == "1h" and lev in ["6", "+6"]:
                lev = lev.replace("6", "5")

            kind = kind.lower()
        else:
            kind = None

        func_choice = self.__make_func_choice(indicator="hit_ratio")

        return func_choice[lev](
            self.observation, self.forecast, kind=kind, lev=int(lev)
        )

    def calc_miss_ratio(self, kind: str = None, lev: str = "0"):
        """Calculste precipitation missing ratio.

        Args:
            kind (str): Statistic kind. Options are '1h', '3h', '12h' and '24h'.
            lev (str): Level ID, Options:
                       * "0": Rain or shine.
                       * "1": Light rain
                       * "2": Moderate rain
                       * "3": Heavy rain
                       * "4": Violent rain
                       * "5": Severe violent rain
                       * "6": Super severe violent rain(1h don't have this)
                       * "+1": Accumulated to Light rain
                       * "+2": Accumulated to Moderate rain
                       * "+3": Accumulated to Heavy rain
                       * "+4": Accumulated to Violent rain
                       * "+5": Accumulated to Severe violent rain
                       * "+6": Accumulated to Super severe violent rain(1h don't have this)
                       Defaults to "0".

        Returns:
            float: Missing ratio of specific kind and level(%)
        """
        if int(lev) > 0:
            if not kind:
                kind = self.kind
            assert kind and (kind.lower() in ["1h", "3h", "12h", "24h"])
            if not lev:
                lev = self.lev
            assert lev and (lev in self.lev_index)

            if kind == "1h" and lev in ["6", "+6"]:
                lev = lev.replace("6", "5")

            kind = kind.lower()
        else:
            kind = None

        func_choice = self.__make_func_choice(indicator="miss_ratio")

        return func_choice[lev](
            self.observation, self.forecast, kind=kind, lev=int(lev)
        )

    def calc_false_alarm_ratio(self, kind: str = None, lev: str = "0"):
        """Calculste precipitation false alarm ratio.

        Args:
            kind (str): Statistic kind. Options are '1h', '3h', '12h' and '24h'.
            lev (str): Level ID, Options:
                       * "0": Rain or shine.
                       * "1": Light rain.
                       * "2": Moderate rain.
                       * "3": Heavy rain.
                       * "4": Violent rain.
                       * "5": Severe violent rain.
                       * "6": Super severe violent rain(1h don't have this).
                       * "+1": Accumulated to Light rain.
                       * "+2": Accumulated to Moderate rain.
                       * "+3": Accumulated to Heavy rain.
                       * "+4": Accumulated to Violent rain.
                       * "+5": Accumulated to Severe violent rain.
                       * "+6": Accumulated to Super severe violent rain(1h don't have this).
                       Defaults to "0".

        Returns:
            float: False alarm ratio of specific kind and level(%)
        """
        if int(lev) > 0:
            if not kind:
                kind = self.kind
            assert kind and (kind.lower() in ["1h", "3h", "12h", "24h"])
            if not lev:
                lev = self.lev
            assert lev and (lev in self.lev_index)

            if kind == "1h" and lev in ["6", "+6"]:
                lev = lev.replace("6", "5")

            kind = kind.lower()
        else:
            kind = None

        func_choice = self.__make_func_choice(indicator="false_alarm_ratio")

        return func_choice[lev](
            self.observation, self.forecast, kind=kind, lev=int(lev)
        )

    def calc_false_alarm_rate(self, kind: str = None, lev: str = "0"):
        """Calculste precipitation false alarm rate.

        Args:
            kind (str): Statistic kind. Options are '1h', '3h', '12h' and '24h'.
            lev (str): Level ID, Options:
                       * "0": Rain or shine.
                       * "1": Light rain.
                       * "2": Moderate rain.
                       * "3": Heavy rain.
                       * "4": Violent rain.
                       * "5": Severe violent rain.
                       * "6": Super severe violent rain(1h don't have this).
                       * "+1": Accumulated to Light rain.
                       * "+2": Accumulated to Moderate rain.
                       * "+3": Accumulated to Heavy rain.
                       * "+4": Accumulated to Violent rain.
                       * "+5": Accumulated to Severe violent rain.
                       * "+6": Accumulated to Super severe violent rain(1h don't have this).
                       Defaults to "0".

        Returns:
            float: False alarm ratio of specific kind and level(%)
        """
        if int(lev) > 0:
            if not kind:
                kind = self.kind
            assert kind and (kind.lower() in ["1h", "3h", "12h", "24h"])
            if not lev:
                lev = self.lev
            assert lev and (lev in self.lev_index)

            if kind == "1h" and lev in ["6", "+6"]:
                lev = lev.replace("6", "5")

            kind = kind.lower()
        else:
            kind = None

        func_choice = self.__make_func_choice(indicator="false_alarm_rate")

        return func_choice[lev](
            self.observation, self.forecast, kind=kind, lev=int(lev)
        )

    def calc_ts(self, kind: str = "1h", lev: str = "0"):
        """Calculste precipitation threat score(TS).

        Args:
            kind (str): Statistic kind. Options are '1h', '3h', '12h' and '24h'.
            lev (str): Level ID, Options:
                       * "0": Rain or shine.
                       * "1": Light rain.
                       * "2": Moderate rain.
                       * "3": Heavy rain.
                       * "4": Violent rain.
                       * "5": Severe violent rain.
                       * "6": Super severe violent rain(1h don't have this).
                       * "+1": Accumulated to Light rain.
                       * "+2": Accumulated to Moderate rain.
                       * "+3": Accumulated to Heavy rain.
                       * "+4": Accumulated to Violent rain.
                       * "+5": Accumulated to Severe violent rain.
                       * "+6": Accumulated to Super severe violent rain(1h don't have this).
                       Defaults to "0".

        Returns:
            float: Threat score(TS) of specific kind and level.
        """
        if int(lev) > 0:
            if not kind:
                kind = self.kind
            assert kind and (kind.lower() in ["1h", "3h", "12h", "24h"])
            if not lev:
                lev = self.lev
            assert lev and (lev in self.lev_index)

            if kind == "1h" and lev in ["6", "+6"]:
                lev = lev.replace("6", "5")

            kind = kind.lower()
        else:
            kind = None

        func_choice = self.__make_func_choice(indicator="ts")

        return func_choice[lev](
            self.observation, self.forecast, kind=kind, lev=int(lev)
        )

    def calc_ets(self, kind: str = None, lev: str = "0"):
        """Calculste precipitation equitable threat score(ETS).

        Args:
            kind (str): Statistic kind. Options are '1h', '3h', '12h' and '24h'.
            lev (str): Level ID, Options:
                       * "0": Rain or shine.
                       * "1": Light rain.
                       * "2": Moderate rain.
                       * "3": Heavy rain.
                       * "4": Violent rain.
                       * "5": Severe violent rain.
                       * "6": Super severe violent rain(1h don't have this).
                       * "+1": Accumulated to Light rain.
                       * "+2": Accumulated to Moderate rain.
                       * "+3": Accumulated to Heavy rain.
                       * "+4": Accumulated to Violent rain.
                       * "+5": Accumulated to Severe violent rain.
                       * "+6": Accumulated to Super severe violent rain(1h don't have this).
                       Defaults to "0".

        Returns:
            float: Equitable threat score(ETS) of specific kind and level
        """
        if int(lev) > 0:
            if not kind:
                kind = self.kind
            assert kind and (kind.lower() in ["1h", "3h", "12h", "24h"])
            if not lev:
                lev = self.lev
            assert lev and (lev in self.lev_index)

            if kind == "1h" and lev in ["6", "+6"]:
                lev = lev.replace("6", "5")

            kind = kind.lower()
        else:
            kind = None

        func_choice = self.__make_func_choice(indicator="ets")

        return func_choice[lev](
            self.observation, self.forecast, kind=kind, lev=int(lev)
        )

    def calc_bias_score(self, kind: str = None, lev: str = "0"):
        """Calculste precipitation equitable threat score(ETS).

        Args:
            kind (str): Statistic kind. Options are '1h', '3h', '12h' and '24h'.
            lev (str): Level ID, Options:
                       * "0": Rain or shine.
                       * "1": Light rain.
                       * "2": Moderate rain.
                       * "3": Heavy rain.
                       * "4": Violent rain.
                       * "5": Severe violent rain.
                       * "6": Super severe violent rain(1h don't have this).
                       * "+1": Accumulated to Light rain.
                       * "+2": Accumulated to Moderate rain.
                       * "+3": Accumulated to Heavy rain.
                       * "+4": Accumulated to Violent rain.
                       * "+5": Accumulated to Severe violent rain.
                       * "+6": Accumulated to Super severe violent rain(1h don't have this).
                       Defaults to "0".

        Returns:
            float: Equitable threat score(ETS) of specific kind and level
        """
        if int(lev) > 0:
            if not kind:
                kind = self.kind
            assert kind and (kind.lower() in ["1h", "3h", "12h", "24h"])
            if not lev:
                lev = self.lev
            assert lev and (lev in self.lev_index)

            if kind == "1h" and lev in ["6", "+6"]:
                lev = lev.replace("6", "5")

            kind = kind.lower()
        else:
            kind = None

        func_choice = self.__make_func_choice(indicator="bias_score")

        return func_choice[lev](
            self.observation, self.forecast, kind=kind, lev=int(lev)
        )

    def gather_all_indicators(self, kind: str = None):
        """Calculate all indicators.

        Args:
            kind (str): Statistic kind. Options are '1h', '3h', '12h' and '24h'.

        Returns:
            pandas.core.frame.DataFrame: All indicators.
        """
        if not kind:
            kind = self.kind
        assert kind and (kind.lower() in ["1h", "3h", "12h", "24h"])

        LEVS = {
            "0": ("rain_or_shine", None),
            "1": ("level", 1),
            "2": ("level", 2),
            "3": ("level", 3),
            "4": ("level", 4),
            "5": ("level", 5),
            "6": ("level", 6),
            "+1": ("accumulative", 1),
            "+2": ("accumulative", 2),
            "+3": ("accumulative", 3),
            "+4": ("accumulative", 4),
            "+5": ("accumulative", 5),
            "+6": ("accumulative", 6),
        }
        KINDS = {"1h": "01_hour", "3h": "03_hours", "24h": "24_hours"}
        funcs = {
            # tuple: (function object, unit, comment)
            "bias_score": (self.calc_bias_score, None, "BIAS score"),
            "ts": (self.calc_ts, None, None),
            "ets": (self.calc_ets, None, None),
            "false_alarm_ratio": (self.calc_false_alarm_ratio, "%", None),
            "missing_ratio": (self.calc_miss_ratio, "%", None),
            "accuracy_ratio": (self.calc_accuracy_ratio, "%", None),
        }
        result = []
        kind_desc = KINDS[kind]

        for lev, (lev_kind, lev_desc) in LEVS.items():
            for func_desc, (func, units, comment) in funcs.items():
                result.append(
                    {
                        "element": "precipitation",
                        "interval": kind_desc,
                        "kind": lev_kind,
                        "level": lev_desc,
                        "indicator": func_desc,
                        "value": func(kind=kind, lev=lev),
                        "unit": units,
                        "comment": comment,
                    }
                )

        df = pd.DataFrame(result)

        return df


if __name__ == "__main__":
    temp = PrecipitationComparison(
        np.array([1, 2, 3, 4]), np.array([1, 2, 3, 9]), unit="m"
    )
    print(temp)
    # temp.gather_all_factors().to_csv("./test.csv")



import numpy as np
from cyeva import identify_direction


def test_iss33():
    from cyeva import WindComparison

    ws1 = np.array([1, 2])
    ws2 = np.array([2, 3])
    wd1 = np.array([357, 357])
    wd2 = np.array([3, 3])

    wc = WindComparison(ws1, ws2, wd1, wd2)
    assert np.isclose(wc.calc_rmse(kind='direction'), 6)


def test_iss36():
    assert (
        np.array([0, 2, 4, 6, 0]) == identify_direction([0, 90, 180, 270, 360])
    ).all()
    assert (
        np.array([0, 4, 8, 12, 0])
        == identify_direction([0, 90, 180, 270, 360], dnum=16)
    ).all()
